from typing import TYPE_CHECKING

from kleur import AltColors, Color, Colored, Colors, ColorTheme, c

from .utils import (
    ArgsParser,
    check_integer_in_range,
    get_class_vars,
    parse_key_value_pair,
    print_lines,
    try_convert,
)

if TYPE_CHECKING:
    from argparse import Namespace
    from collections.abc import Iterable, Iterator


def _color_shade(color: Color) -> str:
    return f"{Colored(color.as_hex.center(8), color.contrasting_shade, color)}"


def _color_line(color: Color, name: str, shades: Iterable[float]) -> str:
    shades_str = "".join(_color_shade(color.shade(s)) for s in shades)
    hue = "   " if name == "grey" else f"{color.hue * 360:03.0f}"
    return f"{shades_str}{Colored(f' {hue} {name}', color)}"


def _color_lines(
    color_theme: dict[str, Color], n_shades: int, n_sats: int
) -> Iterator[str]:
    n = n_shades + 1
    shades = [i / n for i in range(1, n)]

    # Any color will be grey when saturation is set to 0.
    yield _color_line(ColorTheme.grey, "grey", shades)

    # Shade/hue tables for specific saturation values
    for i in range(1, n_sats + 1):
        # Shade percentages row
        yield "".join(f"{s:.2%}".center(8) for s in shades)
        # Tables
        saturation = i / n_sats
        for name, color in sorted(color_theme.items(), key=lambda p: p[1].hue):
            yield _color_line(color.adjust(saturation=saturation), name, shades)


class ThemeArgsParser(ArgsParser):
    name = "theme"

    def _parse_args(self) -> None:
        self._parser.add_argument(
            "-c",
            "--colors",
            nargs="+",
            metavar="NAME=HUE (1-360)",
            type=parse_key_value_pair,
            default={},
        )
        self._parser.add_argument(
            "-m", "--merge-with-default-theme", action="store_true", default=False
        )
        self._parser.add_argument(
            "-a", "--alt-default-theme", action="store_true", default=False
        )
        self._parser.add_argument(
            "-s", "--number-of-shades", type=check_integer_in_range(1, 99), default=9
        )
        self._parser.add_argument(
            "-v", "--number-of-vibrances", type=check_integer_in_range(1, 99), default=2
        )

    def _run_command(self, args: Namespace) -> None:
        hues = {k: c(try_convert(int, h, default=333)) for k, h in args.colors}
        if args.merge_with_default_theme or not hues:
            theme_cls = AltColors if args.alt_default_theme else Colors
            hues = dict(get_class_vars(theme_cls, Color)) | hues
        print_lines(_color_lines(hues, args.number_of_shades, args.number_of_vibrances))



import sys
from abc import ABC, abstractmethod
from argparse import ArgumentParser
from typing import TYPE_CHECKING, ClassVar, overload

if TYPE_CHECKING:
    from argparse import Namespace
    from collections.abc import Callable, Iterable, Iterator


def check_integer(v: str, *, conditions: Callable[[int], bool] = None) -> int:
    value = int(v)
    if conditions and not conditions(value):
        raise ValueError(value)
    return value


def check_integer_in_range(low: int | None, high: int | None) -> Callable[[str], int]:
    def is_in_range(n: int) -> bool:
        return (low is None or n >= low) and (high is None or n <= high)

    def check(v: str) -> int:
        return check_integer(v, conditions=is_in_range)

    return check


def parse_key_value_pair(value: str) -> tuple[str, str]:
    key, value = value.split("=", 1)
    return key, value


def try_parse_key_value_pair(value: str) -> str | tuple[str, str]:
    try:
        return parse_key_value_pair(value)
    except ValueError:
        return value


def print_lines(lines: Iterable[str]) -> None:
    for line in lines:
        sys.stdout.write(f"{line}\n")


@overload
def try_convert[T, R](cls: Callable[[T], R], val: T, *, default: R) -> R: ...


@overload
def try_convert[T, R](
    cls: Callable[[T], R], val: T, *, default: None = None
) -> R | None: ...


def try_convert[T, R](cls: Callable[[T], R], val: T, *, default: R = None) -> R | None:
    try:
        return cls(val)
    except ValueError:
        return default


def get_class_vars[T](cls: type, value_type: type[T]) -> Iterator[tuple[str, T]]:
    for k, c in cls.__dict__.items():
        if not k.startswith("_") and isinstance(c, value_type):
            yield k, c


class ArgsParser(ABC):
    name: ClassVar[str]

    def __init__(self, parser: ArgumentParser) -> None:
        self._parser = parser
        self._parse_args()
        parser.set_defaults(func=self._run_command)

    @abstractmethod
    def _parse_args(self) -> None: ...

    @abstractmethod
    def _run_command(self, args: Namespace) -> None: ...


def run_command(*sub_parsers: type[ArgsParser]) -> None:
    parser = ArgumentParser()
    subparsers = parser.add_subparsers(required=True)
    for cls in sub_parsers:
        cls(subparsers.add_parser(cls.name))
    args = parser.parse_args()
    args.func(args)
